/*
 * Hibernate Search, full-text search for your domain model
 *
 * License: GNU Lesser General Public License (LGPL), version 2.1 or later
 * See the lgpl.txt file in the root directory or <http://www.gnu.org/licenses/lgpl-2.1.html>.
 */
package org.hibernate.search.util.impl.integrationtest.common.rule;

import static org.assertj.core.api.Assertions.assertThat;

import java.util.ArrayList;
import java.util.List;

import org.hibernate.search.engine.backend.document.converter.runtime.FromIndexFieldValueConvertContext;
import org.hibernate.search.engine.search.SearchResult;
import org.hibernate.search.engine.search.query.spi.LoadingResult;
import org.hibernate.search.engine.search.query.spi.ProjectionHitMapper;
import org.hibernate.search.util.impl.integrationtest.common.assertion.StubSearchWorkAssert;
import org.hibernate.search.util.impl.integrationtest.common.stub.backend.search.StubSearchWork;
import org.hibernate.search.util.impl.integrationtest.common.stub.backend.search.projection.impl.StubSearchProjection;

class SearchWorkCall<T> {

	private final List<String> indexNames;
	private final StubSearchWork work;
	private final FromIndexFieldValueConvertContext convertContext;
	private final ProjectionHitMapper<?, ?> projectionHitMapper;
	private final StubSearchProjection<T> rootProjection;
	private final StubSearchWorkBehavior<?> behavior;

	SearchWorkCall(List<String> indexNames,
			StubSearchWork work,
			FromIndexFieldValueConvertContext convertContext,
			ProjectionHitMapper<?, ?> projectionHitMapper,
			StubSearchProjection<T> rootProjection) {
		this.indexNames = indexNames;
		this.work = work;
		this.convertContext = convertContext;
		this.projectionHitMapper = projectionHitMapper;
		this.rootProjection = rootProjection;
		this.behavior = null;
	}

	SearchWorkCall(List<String> indexNames,
			StubSearchWork work,
			StubSearchWorkBehavior<?> behavior) {
		this.indexNames = indexNames;
		this.work = work;
		this.convertContext = null;
		this.projectionHitMapper = null;
		this.rootProjection = null;
		this.behavior = behavior;
	}

	public <U> SearchResult<U> verify(SearchWorkCall<U> actualCall) {
		assertThat( actualCall.indexNames )
				.as( "Search work did not target the expected indexes: " )
				.isEqualTo( indexNames );
		StubSearchWorkAssert.assertThat( actualCall.work )
				.as( "Search work on indexes " + indexNames + " did not match: " )
				.matches( work );

		long totalHitCount = behavior.getTotalHitCount();

		return new SearchResultFromMock<>( totalHitCount, getResults( actualCall.convertContext,
				actualCall.projectionHitMapper, actualCall.rootProjection, behavior.getRawHits() ) );
	}

	private static <U> List<U> getResults(FromIndexFieldValueConvertContext actualConvertContext,
			ProjectionHitMapper<?, ?> actualProjectionHitMapper,
			StubSearchProjection<U> actualRootProjection,
			List<?> rawHits) {
		List<Object> extractedElements = new ArrayList<>( rawHits.size() );

		for ( Object rawHit : rawHits ) {
			extractedElements.add(
					actualRootProjection.extract( actualProjectionHitMapper, rawHit, actualConvertContext ) );
		}

		LoadingResult<?> loadingResult = actualProjectionHitMapper.load();

		List<U> results = new ArrayList<>( rawHits.size() );

		for ( Object extractedElement : extractedElements ) {
			results.add( (U) actualRootProjection.transform( loadingResult, extractedElement ) );
		}

		return results;
	}

	@Override
	public String toString() {
		return "search work execution on indexes '" + indexNames + "'; work = " + work;
	}

	private static final class SearchResultFromMock<T> implements SearchResult<T> {

		private final long totalHitCount;
		private final List<T> hits;

		public SearchResultFromMock(long totalHitCount, List<T> hits) {
			this.totalHitCount = totalHitCount;
			this.hits = hits;
		}

		@Override
		public long getHitCount() {
			return totalHitCount;
		}

		@Override
		public List<T> getHits() {
			return hits;
		}

		@Override
		public String toString() {
			return getClass().getSimpleName() + "{" +
					"totalHitCount=" + totalHitCount +
					", hits=" + hits +
					'}';
		}
	}

}
